from typing import Callable

from django.conf import settings
from djmoney.money import Money

from django_ecommerce.models import ProductVariant, Setting, Order, OrderItem, City, Product, Category, Country


def chunks(queryset, chunk_size):
    chunk_size = int(chunk_size)

    return [queryset[i:i + chunk_size] for i in range(0, len(queryset), chunk_size)]


def similar_products(product, limit=4):
    category = product.category
    similar_items = Product.objects.filter(category=category).exclude(id=product.id)

    return similar_items.order_by('?')[:limit]


def variant_availability(variant: ProductVariant, request) -> int:
    cart_quantity = 0

    if str(variant.pk) in request.session.get('cart', {}):
        cart_quantity = request.session['cart'][str(variant.pk)]

    return variant.available_quantity - cart_quantity


def money(amount: float, currency: str = None) -> Money:
    if currency:
        return Money(amount, currency)

    if hasattr(settings, 'ECOMMERCE_DEFAULT_CURRENCY'):
        return Money(amount, getattr(settings, 'ECOMMERCE_DEFAULT_CURRENCY'))

    return Money(amount, 'EUR')


def get_shipping_data(amount: Money) -> dict:
    site_settings = Setting.get_solo()

    shipping = {
        'shipping_enabled': site_settings.shipping_enabled,
        'free_shipping_enabled': site_settings.free_shipping_enabled,
        'shipping_amount': site_settings.shipping_amount
    }

    if site_settings.shipping_enabled:
        if site_settings.free_shipping_enabled:
            shipping['free_shipping_remaining'] = site_settings.free_shipping_min_amount - amount
            shipping['is_free_shipping'] = (shipping['free_shipping_remaining']
                                            <= Money(0, site_settings.free_shipping_min_amount_currency))

            if shipping['is_free_shipping']:
                shipping['shipping_amount'] = money(0)

    return shipping


def get_cart(request):
    cart_total = money(0)
    cart_quantity = 0

    items = []
    remove = []

    for variant_id, quantity in request.session.get('cart', {}).items():
        try:
            if quantity <= 0:
                remove.append(variant_id)
                continue

            variant = ProductVariant.objects.get(pk=variant_id)

            if variant.available_quantity <= 0:
                remove.append(variant_id)
                continue

            if quantity > variant.available_quantity:
                quantity = variant.available_quantity

            total = variant.product.actual_price * quantity
            cart_total += total
            cart_quantity += quantity

            items.append({
                'variant': variant,
                'quantity': quantity,
                'total': total
            })
        except ProductVariant.DoesNotExist:
            continue

    if remove:
        for variant_id in remove:
            del request.session['cart'][variant_id]
            continue

        request.session.modified = True

    shipping = get_shipping_data(cart_total)

    if shipping['shipping_enabled'] and shipping['shipping_amount']:
        cart_total += shipping['shipping_amount']

    return {
        'items': items,
        'total': cart_total,
        'quantity': cart_quantity,
        'shipping': shipping,
    }


def remove_from_cart(request, variant_id: int) -> dict:
    if 'cart' in request.session and str(variant_id) in request.session['cart']:
        del request.session['cart'][str(variant_id)]
        request.session.modified = True

    return {}


def add_to_cart(request, variant_id: int, quantity: int = 1) -> dict:
    if variant_id and quantity:
        variant_id = int(variant_id or 0)
        quantity = int(quantity or 0)
        session_key = str(variant_id)

        try:
            variant = ProductVariant.objects.get(pk=variant_id)
        except ProductVariant.DoesNotExist:
            return {'error': f'ProductVariantDoesNotExist'}

        if 'cart' not in request.session:
            request.session['cart'] = {}

        if session_key in request.session['cart']:
            total_quantity = quantity + request.session['cart'][session_key]

            if variant.available_quantity < total_quantity:
                return {'error': f'StockError'}

            remaining = (variant.available_quantity - total_quantity) > 0

            if total_quantity <= 0:
                del request.session['cart'][session_key]
                request.session.modified = True
                return {'remaining': remaining}

            request.session['cart'][session_key] += quantity
            request.session.modified = True

            return {'remaining': remaining}

        if variant.available_quantity < quantity:
            return {'error': f'StockError'}

        remaining = (variant.available_quantity - quantity) > 0

        request.session['cart'][session_key] = quantity
        request.session.modified = True

        return {'remaining': remaining}

    return {'error': 'InvalidRequestOrParameters'}


def create_order(request, email: str = None, full_name: str = None, address: str = None, phone: str = None,
                 city: str = None, country: str = None, zip_code: str = None, remark: str = None,
                 func: Callable = None):
    cart = get_cart(request)

    if not cart['items']:
        return {'error': 'EmptyCart'}

    order = Order.objects.create(
        email=email,
        full_name=full_name,
        address=address,
        phone=phone,
        city=city,
        country=country,
        zip_code=zip_code,
        remark=remark,
        shipping_amount=cart['shipping']['shipping_amount'],
    )

    for item in cart['items']:
        OrderItem.objects.create(
            order=order,
            product_variant=item['variant'],
            price=item['variant'].product.price,
            discount_percentage=item['variant'].product.discount,
            quantity=item['quantity']
        )

    if func:
        func(request, order)

    del request.session['cart']

    return {'order': order}


def get_cities() -> set:
    return {city.name for city in City.objects.all()}


def get_countries() -> set:
    return {country.name for country in Country.objects.all()}


def find_product_by_slug(slug: str) -> Product | None:
    try:
        return Product.objects.get(slug=slug)
    except Product.DoesNotExist:
        return None


def find_category_by_slug(slug: str) -> Category | None:
    try:
        return Category.objects.get(slug=slug)
    except Category.DoesNotExist:
        return None


def find_order_by_code(code: str) -> Order | None:
    try:
        return Order.objects.get(code=code)
    except Order.DoesNotExist:
        return None



from django_ecommerce.models import Setting
from django_ecommerce.utils import get_cart


def ecommerce_settings(request):
    return {
        'ecommerce_settings': Setting.get_solo()
    }


def ecommerce_cart(request):
    return {
        'ecommerce_cart': get_cart(request)
    }



# Generated by Django 5.0.3 on 2024-03-06 10:50

import django_cloud_storage.storage
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ecommerce', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='category',
            name='seo_image_cropped',
            field=models.ImageField(blank=True, editable=False, null=True, storage=django_cloud_storage.storage.CloudStorage, upload_to=''),
        ),
        migrations.AddField(
            model_name='product',
            name='seo_image_cropped',
            field=models.ImageField(blank=True, editable=False, null=True, storage=django_cloud_storage.storage.CloudStorage, upload_to=''),
        ),
        migrations.AddField(
            model_name='productimage',
            name='cropped',
            field=models.ImageField(blank=True, editable=False, null=True, storage=django_cloud_storage.storage.CloudStorage, upload_to=''),
        ),
    ]



from django import template

from django_ecommerce.utils import variant_availability, chunks, similar_products

register = template.Library()

register.filter(chunks)
register.filter(similar_products)
register.filter(variant_availability)



# Generated by Django 5.0.1 on 2024-01-25 09:27

import autoslug.fields
import django.db.models.deletion
import django_cloud_storage.storage
import django_ecommerce.models
import djmoney.models.fields
import django_cloud_thumbnails.fields
from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('slug', autoslug.fields.AutoSlugField(blank=True, editable=True, populate_from='name', unique=True)),
                ('description', models.TextField(blank=True, null=True)),
                ('seo_title', models.CharField(blank=True, max_length=60, null=True)),
                ('seo_description', models.TextField(blank=True, max_length=160, null=True)),
                ('seo_keywords', models.TextField(blank=True, max_length=255, null=True)),
                ('seo_image', django_cloud_thumbnails.fields.ImageCropField(blank=True, null=True, storage=django_cloud_storage.storage.CloudStorage, upload_to='')),
                ('seo_image_cropping', django_cloud_thumbnails.fields.ImageRatioField('seo_image', '1200x630', adapt_rotation=False, allow_full_size=False, free_crop=False, help_text=None, hide_image_field=False, size_warning=False, verbose_name='seo image cropping')),
            ],
            options={
                'verbose_name_plural': 'categories',
            },
        ),
        migrations.CreateModel(
            name='City',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'verbose_name_plural': 'cities',
            },
        ),
        migrations.CreateModel(
            name='Color',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Country',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'verbose_name_plural': 'countries',
            },
        ),
        migrations.CreateModel(
            name='Discount',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('active', models.BooleanField(default=False)),
                ('name', models.CharField(max_length=255)),
                ('start_date', models.DateTimeField()),
                ('end_date', models.DateTimeField()),
            ],
        ),
        migrations.CreateModel(
            name='Gender',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Order',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(blank=True, db_index=True, default=django_ecommerce.models.generate_order_code, max_length=16)),
                ('status', models.CharField(choices=[('pending', 'Pending'), ('confirmed', 'Confirmed'), ('canceled', 'Canceled'), ('sent', 'Sent'), ('returned', 'Returned'), ('delivered', 'Delivered')], default='pending', max_length=16)),
                ('email', models.EmailField(blank=True, max_length=255, null=True)),
                ('full_name', models.CharField(blank=True, max_length=255, null=True)),
                ('address', models.CharField(blank=True, max_length=255, null=True)),
                ('phone', models.CharField(blank=True, max_length=255, null=True)),
                ('city', models.CharField(blank=True, max_length=255, null=True)),
                ('country', models.CharField(blank=True, max_length=255, null=True)),
                ('zip_code', models.CharField(blank=True, max_length=255, null=True)),
                ('remark', models.TextField(blank=True, null=True)),
                ('shipping_amount_currency', djmoney.models.fields.CurrencyField(choices=[('EUR', 'EUR')], default='EUR', editable=False, max_length=3, null=True)),
                ('shipping_amount', djmoney.models.fields.MoneyField(blank=True, currency_choices=[('EUR', 'EUR')], decimal_places=2, default=Decimal('0'), default_currency='EUR', max_digits=14, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='Setting',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('shipping_enabled', models.BooleanField(default=False)),
                ('shipping_amount_currency', djmoney.models.fields.CurrencyField(choices=[('EUR', 'EUR')], default='EUR', editable=False, max_length=3, null=True)),
                ('shipping_amount', djmoney.models.fields.MoneyField(blank=True, currency_choices=[('EUR', 'EUR')], decimal_places=2, default=Decimal('0'), default_currency='EUR', max_digits=14, null=True)),
                ('free_shipping_enabled', models.BooleanField(default=False)),
                ('free_shipping_min_amount_currency', djmoney.models.fields.CurrencyField(choices=[('EUR', 'EUR')], default='EUR', editable=False, max_length=3, null=True)),
                ('free_shipping_min_amount', djmoney.models.fields.MoneyField(blank=True, currency_choices=[('EUR', 'EUR')], decimal_places=2, default=Decimal('0'), default_currency='EUR', max_digits=14, null=True)),
            ],
            options={
                'verbose_name': 'Settings',
            },
        ),
        migrations.CreateModel(
            name='StockImport',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('comment', models.TextField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='Warehouse',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='DiscountCategory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('percentage', models.PositiveIntegerField(default=0)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='discount_categories', to='ecommerce.category')),
                ('discount', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='categories', to='ecommerce.discount')),
            ],
            options={
                'verbose_name_plural': 'discount categories',
            },
        ),
        migrations.CreateModel(
            name='Product',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('slug', autoslug.fields.AutoSlugField(blank=True, editable=True, populate_from='name', unique=True)),
                ('description', models.TextField(blank=True, null=True)),
                ('price_currency', djmoney.models.fields.CurrencyField(choices=[('EUR', 'EUR')], default='EUR', editable=False, max_length=3)),
                ('price', djmoney.models.fields.MoneyField(currency_choices=[('EUR', 'EUR')], decimal_places=2, default_currency='EUR', max_digits=14)),
                ('seo_title', models.CharField(blank=True, max_length=60, null=True)),
                ('seo_description', models.TextField(blank=True, max_length=160, null=True)),
                ('seo_keywords', models.TextField(blank=True, max_length=255, null=True)),
                ('seo_image', django_cloud_thumbnails.fields.ImageCropField(blank=True, null=True, storage=django_cloud_storage.storage.CloudStorage, upload_to='')),
                ('seo_image_cropping', django_cloud_thumbnails.fields.ImageRatioField('seo_image', '1200x630', adapt_rotation=False, allow_full_size=False, free_crop=False, help_text=None, hide_image_field=False, size_warning=False, verbose_name='seo image cropping')),
                ('category', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='products', to='ecommerce.category')),
                ('color', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='ecommerce.color')),
                ('gender', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='ecommerce.gender')),
            ],
        ),
        migrations.CreateModel(
            name='DiscountProduct',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('percentage', models.PositiveIntegerField(default=0)),
                ('discount', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='products', to='ecommerce.discount')),
                ('product', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='discount_products', to='ecommerce.product')),
            ],
        ),
        migrations.CreateModel(
            name='ProductImage',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image', django_cloud_thumbnails.fields.ImageCropField(blank=True, null=True, storage=django_cloud_storage.storage.CloudStorage, upload_to='')),
                ('cropping', django_cloud_thumbnails.fields.ImageRatioField('image', '1000x1000', adapt_rotation=False, allow_full_size=False, free_crop=False, help_text=None, hide_image_field=False, size_warning=False, verbose_name='cropping')),
                ('position', models.PositiveIntegerField(default=0)),
                ('product', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='images', to='ecommerce.product')),
            ],
            options={
                'ordering': ['position'],
            },
        ),
        migrations.CreateModel(
            name='ProductVariant',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sku', models.CharField(max_length=255, unique=True)),
                ('name', models.CharField(max_length=255)),
                ('position', models.PositiveIntegerField(default=0)),
                ('active', models.BooleanField(default=True)),
                ('product', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='variants', to='ecommerce.product')),
            ],
        ),
        migrations.CreateModel(
            name='OrderItem',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('price_currency', djmoney.models.fields.CurrencyField(choices=[('EUR', 'EUR')], default='EUR', editable=False, max_length=3)),
                ('price', djmoney.models.fields.MoneyField(currency_choices=[('EUR', 'EUR')], decimal_places=2, default_currency='EUR', max_digits=14)),
                ('discount_percentage', models.PositiveIntegerField(blank=True, default=0, null=True)),
                ('quantity', models.PositiveIntegerField()),
                ('reserved', models.BooleanField(default=False)),
                ('order', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='items', to='ecommerce.order')),
                ('product_variant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='order_items', to='ecommerce.productvariant')),
                ('warehouse', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='order_items', to='ecommerce.warehouse')),
            ],
        ),
        migrations.CreateModel(
            name='Stock',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.IntegerField()),
                ('object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('product_variant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='stock', to='ecommerce.productvariant')),
                ('warehouse', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='stock', to='ecommerce.warehouse')),
            ],
            options={
                'verbose_name_plural': 'stock',
                'indexes': [models.Index(fields=['content_type', 'object_id'], name='ecommerce_s_content_aff7e3_idx')],
            },
        ),
    ]



from unittest.mock import MagicMock

from django.test import TestCase
from django.test.utils import override_settings
from djmoney.money import Money

from django_ecommerce.models import Category, Product, Color, Gender, ProductVariant, StockImport, Warehouse, Stock, \
    Country, City, Order, Setting
from django_ecommerce.utils import chunks, money, similar_products, variant_availability, add_to_cart, create_order, \
    find_order_by_code, find_category_by_slug, find_product_by_slug, get_countries, get_cities, remove_from_cart, \
    get_cart, get_shipping_data


class UtilsTestCase(TestCase):
    def setUp(self):
        category = Category.objects.create(name='Test')
        color = Color.objects.create(name='Blue')
        gender = Gender.objects.create(name='Unisex')
        Warehouse.objects.create(name='Warehouse1')

        p = Product.objects.create(name='Product1', category=category, color=color, gender=gender, price=money(10))
        ProductVariant.objects.create(product=p, name='Small', sku='PRODUCT1-SMALL')
        ProductVariant.objects.create(product=p, name='Medium', sku='PRODUCT1-MEDIUM')

        Product.objects.create(name='Product2', category=category, color=color, gender=gender, price=money(15))
        Product.objects.create(name='Product3', category=category, color=color, gender=gender, price=money(20))

    def test_chunks(self):
        queryset = [1, 2, 3, 4, 5, 6, 7, 8]
        chunk_size = 3
        result = chunks(queryset, chunk_size)

        self.assertEqual(result, [[1, 2, 3], [4, 5, 6], [7, 8]])

    def test_similar_products(self):
        similar = similar_products(Product.objects.first())
        self.assertTrue(similar)

    def test_variant_availability(self):
        product = Product.objects.first()
        request = self.client.get('/').wsgi_request

        # No stock
        self.assertEqual(variant_availability(product.variants.first(), request), 0)

        # Add to stock
        si = StockImport.objects.create()
        Stock.objects.create(product_variant=product.variants.first(), warehouse=Warehouse.objects.first(), quantity=5,
                             content_object=si)
        self.assertEqual(variant_availability(product.variants.first(), request), 5)

        # Add item to cart
        add_to_cart(request, product.variants.first().pk, 1)
        self.assertEqual(variant_availability(product.variants.first(), request), 4)

        # Create order and reserve items
        order = create_order(request)['order']
        for item in order.items.all():
            item.reserved = True
            item.save()

        self.assertEqual(variant_availability(product.variants.first(), request), 4)

    def test_money(self):
        # Currency argument is provided
        self.assertEqual(money(10, 'USD'), Money(10, 'USD'))

        # Currency argument is not provided, and we have default currency in settings
        with override_settings(ECOMMERCE_DEFAULT_CURRENCY='EUR'):
            self.assertEqual(money(10), Money(10, 'EUR'))

        # Default value for currency
        self.assertEqual(money(10), Money(10, 'EUR'))

    def test_get_shipping_data(self):
        product = Product.objects.first()

        # Shipping is not enabled
        shipping = get_shipping_data(product.price)
        self.assertFalse(shipping['shipping_enabled'])
        self.assertEqual(shipping['shipping_amount'], money(0))

        # Enable shipping
        ecommerce_settings = Setting.get_solo()
        ecommerce_settings.shipping_enabled = True
        ecommerce_settings.shipping_amount = money(3)
        ecommerce_settings.save()

        shipping = get_shipping_data(product.price)
        self.assertTrue(shipping['shipping_enabled'])
        self.assertEqual(shipping['shipping_amount'], money(3))

        # Enable free shipping
        ecommerce_settings.free_shipping_enabled = True
        ecommerce_settings.free_shipping_min_amount = 3 * product.price
        ecommerce_settings.save()

        shipping = get_shipping_data(2 * product.price)
        self.assertTrue(shipping['shipping_enabled'])
        self.assertTrue(shipping['free_shipping_enabled'])
        self.assertEqual(shipping['shipping_amount'], money(3))

        shipping = get_shipping_data(3 * product.price)
        self.assertTrue(shipping['shipping_enabled'])
        self.assertEqual(shipping['shipping_amount'], money(0))

    def test_get_cart(self):
        product = Product.objects.first()
        request = self.client.get('/').wsgi_request

        # Empty Cart
        cart = get_cart(request)
        self.assertEqual(cart['total'], money(0))
        self.assertEqual(cart['quantity'], 0)

        # Add item to cart and check again
        si = StockImport.objects.create()
        stock = Stock.objects.create(product_variant=product.variants.first(), warehouse=Warehouse.objects.first(),
                                     quantity=5,
                                     content_object=si)

        self.assertEqual(add_to_cart(request, product.variants.first().pk, 2), {'remaining': True})

        cart = get_cart(request)
        self.assertEqual(cart['total'], 2 * product.price)
        self.assertEqual(cart['quantity'], 2)

        # Check if quantity is updated when stock is updated
        stock.quantity = 1
        stock.save()
        cart = get_cart(request)
        self.assertEqual(cart['total'], product.price)
        self.assertEqual(cart['quantity'], 1)

        # Check if item is removed from cart
        stock.delete()
        cart = get_cart(request)
        self.assertEqual(cart['total'], money(0))
        self.assertEqual(cart['quantity'], 0)

    def test_remove_from_cart(self):
        product = Product.objects.first()
        request = self.client.get('/').wsgi_request

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=product.variants.first(), warehouse=Warehouse.objects.first(), quantity=5,
                             content_object=si)

        self.assertEqual(add_to_cart(request, product.variants.first().pk, 3), {'remaining': True})
        remove_from_cart(request, product.variants.first().pk)
        self.assertNotIn(str(product.variants.first()), request.session.get('cart'))

    def test_add_to_cart(self):
        product = Product.objects.first()
        request = self.client.get('/').wsgi_request

        # Check invalid variant id
        self.assertEqual(add_to_cart(request, 12345), {'error': f'ProductVariantDoesNotExist'})

        # Check for stock data
        self.assertEqual(add_to_cart(request, product.variants.first().pk), {'error': f'StockError'})

        # Add stock and then add to cart
        si = StockImport.objects.create()
        Stock.objects.create(product_variant=product.variants.first(), warehouse=Warehouse.objects.first(), quantity=5,
                             content_object=si)

        self.assertEqual(add_to_cart(request, product.variants.first().pk, 4), {'remaining': True})
        self.assertEqual(add_to_cart(request, product.variants.first().pk, 1), {'remaining': False})
        self.assertEqual(add_to_cart(request, product.variants.first().pk, -1), {'remaining': True})

        # Check if it was removed from cart correctly
        add_to_cart(request, product.variants.first().pk, -4)
        self.assertNotIn(str(product.variants.first()), request.session.get('cart'))

    def test_create_order(self):
        product = Product.objects.first()
        request = self.client.get('/').wsgi_request

        # Create a mock function
        mock_func = MagicMock()

        # Order params
        params = {
            'email': 'test@example.com',
            'full_name': 'Hello World',
            'address': 'Test address',
            'phone': '1234567890',
            'city': 'Test',
            'country': 'Test',
            'zip_code': '123456',
            'remark': 'Test',
        }

        # Check when cart is empty
        self.assertEqual(create_order(request), {'error': 'EmptyCart'})

        # Add item to cart and create order
        si = StockImport.objects.create()
        Stock.objects.create(product_variant=product.variants.first(), warehouse=Warehouse.objects.first(), quantity=5,
                             content_object=si)

        add_to_cart(request, product.variants.first().pk, 2)
        result = create_order(request, func=mock_func, **params)

        # Assert that the order was created
        self.assertTrue('order' in result)
        order = result['order']
        self.assertIsInstance(order, Order)

        # Assert that the session cart is empty after order creation
        self.assertNotIn('cart', request)

        # Check if params were saved correctly
        for k, v in params.items():
            self.assertEqual(v, getattr(order, k))

        # Assert that the mock function was called with the expected arguments
        mock_func.assert_called_once_with(request, order)

        # Check if order item was saved correctly
        item = order.items.first()
        self.assertEqual(item.product_variant, product.variants.first())
        self.assertEqual(item.price, product.price)
        self.assertEqual(item.quantity, 2)

    def test_get_cities(self):
        c1 = City.objects.create(name='City1')
        c2 = City.objects.create(name='City1')

        self.assertEqual(get_cities(), {c1.name, c2.name})

    def test_get_countries(self):
        c1 = Country.objects.create(name='Country1')
        c2 = Country.objects.create(name='Country2')

        self.assertEqual(get_countries(), {c1.name, c2.name})

    def test_find_product_by_slug(self):
        p = Product.objects.first()

        self.assertEqual(find_product_by_slug(p.slug), p)

    def test_find_category_by_slug(self):
        c = Category.objects.create(name='MyCategory', slug='my-category')

        self.assertEqual(find_category_by_slug(c.slug), c)

    def test_find_order_by_code(self):
        product = Product.objects.first()

        request = self.client.get('/').wsgi_request

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=product.variants.first(), warehouse=Warehouse.objects.first(), quantity=5,
                             content_object=si)

        add_to_cart(request, product.variants.first().pk)

        order = create_order(request)['order']

        self.assertEqual(find_order_by_code(order.code), order)



from adminsortable2.admin import SortableTabularInline, SortableAdminBase
from django import forms
from django.conf import settings
from django.contrib import admin
from django.contrib.contenttypes.admin import GenericTabularInline
from django.forms import ModelForm
from django_cloud_thumbnails import ImageCroppingMixin
from solo.admin import SingletonModelAdmin

from .models import ProductImage, Gender, Category, Color, Product, ProductVariant, Stock, Warehouse, Setting, Order, \
    OrderItem, StockImport, City, DiscountCategory, DiscountProduct, Discount, Country


class ProductImageInline(ImageCroppingMixin, SortableTabularInline):
    model = ProductImage
    ordering = ['position']


class AlwaysChangedModelForm(ModelForm):
    def has_changed(self):
        return True


class ProductVariantInline(SortableTabularInline):
    model = ProductVariant
    form = AlwaysChangedModelForm

    fields = ('sku', 'name', 'stock_data', 'reservations_data', 'available_quantity', 'active')
    readonly_fields = ['id', 'stock_data', 'reservations_data', 'available_quantity']
    extra = 0
    ordering = ['position']
    search_fields = ['product']

    def stock_data(self, obj: ProductVariant):
        data = obj.stock_data()

        return ', '.join([f'{x[1]} ({x[0]})' for x in data]) if data else '-'

    stock_data.short_description = 'Stock'

    def reservations_data(self, obj: ProductVariant):
        data = obj.reservations_data()

        return ', '.join([f'{x[1]} ({x[0]})' for x in data]) if data else '-'

    reservations_data.short_description = 'Reservations'

    def available_quantity(self, obj: ProductVariant):
        return obj.available_quantity if obj.available_quantity else '-'

    available_quantity.short_description = 'Available quantity'


@admin.register(ProductVariant)
class ProductVariantAdmin(admin.ModelAdmin):
    fields = ['sku', 'product', 'name', 'active']

    def has_module_permission(self, request):
        return False

    def get_readonly_fields(self, request, obj=None):
        fields = ['id']

        if obj and obj.pk:
            fields += ['sku', 'product', 'name']

        return fields

    search_fields = ['sku', 'name']


@admin.register(Product)
class ProductAdmin(ImageCroppingMixin, SortableAdminBase, admin.ModelAdmin):
    inlines = [ProductImageInline, ProductVariantInline]
    list_display = ('name', 'slug', 'category', 'price')

    search_fields = ['name']
    list_filter = ['category', 'gender', 'color']

    fieldsets = [
        ('General information', {
            'fields': [
                'category', 'color', 'gender', 'price'
            ],
        }),

        ('Web page details', {
            'fields': [
                'slug', 'name', 'description'
            ],
        }),
        ('SEO', {
            'fields': [
                'seo_title', 'seo_description', 'seo_keywords', 'seo_image', 'seo_image_cropping'
            ],
        }),
    ]

    def get_readonly_fields(self, request, obj=None):
        fields = []

        if obj and obj.pk:
            fields.append('sku')

        return fields


@admin.register(Category)
class CategoryAdmin(ImageCroppingMixin, admin.ModelAdmin):
    list_display = ['name']

    fieldsets = [
        ('Web page details', {
            'fields': [
                'slug', 'name', 'description'
            ],
        }),
        ('SEO', {
            'fields': [
                'seo_title', 'seo_description', 'seo_keywords', 'seo_image', 'seo_image_cropping'
            ],
        }),
    ]


@admin.register(Color)
class ColorAdmin(admin.ModelAdmin):
    list_display = ['name']


@admin.register(City)
class CityAdmin(admin.ModelAdmin):
    list_display = ['name']


@admin.register(Country)
class CountryAdmin(admin.ModelAdmin):
    list_display = ['name']


@admin.register(Gender)
class GenderAdmin(admin.ModelAdmin):
    list_display = ['name']


@admin.register(Warehouse)
class WarehouseAdmin(admin.ModelAdmin):
    list_display = ['name']
    search_fields = ['name']


@admin.register(Stock)
class StockAdmin(admin.ModelAdmin):
    list_display = ['product_variant', 'warehouse', 'quantity', 'content_object_link', 'created_at']
    ordering = ['-created_at']
    search_fields = ['product_variant__sku', 'product_variant__name']
    list_filter = ['quantity', 'warehouse']


class StockInline(GenericTabularInline):
    model = Stock
    extra = 0
    autocomplete_fields = ['warehouse', 'product_variant']


@admin.register(StockImport)
class StockImport(admin.ModelAdmin):
    inlines = [StockInline]
    ordering = ['-created_at']
    list_display = ['created_at', 'comment']


@admin.register(Setting)
class SettingAdmin(ImageCroppingMixin, SingletonModelAdmin):
    fieldsets = [
        ('Shipping', {
            'fields': [
                'shipping_enabled', 'shipping_amount', 'free_shipping_enabled', 'free_shipping_min_amount'
            ],
        }),
    ]


class OrderItemInline(admin.TabularInline):
    model = OrderItem
    extra = 0

    def has_change_permission(self, request, obj=None):
        return not obj or obj.can_modify_order_items()

    def has_delete_permission(self, request, obj=None):
        return not obj or obj.can_modify_order_items()

    def has_add_permission(self, request, obj):
        return not obj or obj.can_modify_order_items()

    def get_formset(self, request, obj=None, **kwargs):
        formset = super(OrderItemInline, self).get_formset(request, obj, **kwargs)
        field = formset.form.base_fields['warehouse']
        field.widget.can_add_related = False
        field.widget.can_change_related = False
        field.widget.can_delete_related = False

        return formset


class OrderAdminForm(forms.ModelForm):
    class Meta:
        model = Order
        fields = '__all__'

    city = forms.ModelChoiceField(queryset=City.objects.all(), to_field_name='name', widget=forms.Select,
                                  required=False)
    country = forms.ModelChoiceField(queryset=Country.objects.all(), to_field_name='name', widget=forms.Select,
                                     required=False)


def get_shipping_fields():
    if hasattr(settings, 'ECOMMERCE_SHIPPING_ADMIN_FIELDS'):
        return getattr(settings, 'ECOMMERCE_SHIPPING_ADMIN_FIELDS')

    return ['full_name', 'address', 'city', 'zip_code', 'country', 'phone', 'remark', 'shipping_amount']


@admin.register(Order)
class OrderAdmin(admin.ModelAdmin):
    form = OrderAdminForm
    inlines = [OrderItemInline]
    list_display = ['code', 'status', 'full_name', 'created_at']
    list_filter = ['status']
    search_fields = ['code', 'full_name']
    readonly_fields = ['code', 'total_amount', 'created_at', 'items']

    fieldsets = [
        ('General information', {
            'fields': [
                'code', 'created_at', 'total_amount', 'status', 'email'
            ],
        }),
        ('Shipping', {
            'fields': get_shipping_fields()
        }),
    ]


class DiscountCategoryInline(admin.TabularInline):
    model = DiscountCategory
    extra = 0


class DiscountProductInline(admin.TabularInline):
    model = DiscountProduct
    extra = 0


@admin.register(Discount)
class DiscountAdmin(admin.ModelAdmin):
    inlines = [DiscountCategoryInline, DiscountProductInline]



# Generated by Django on 2024-03-06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "ecommerce",
            "0002_category_seo_image_cropped_product_seo_image_cropped_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="discount",
            name="last_activation_run",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="discount",
            name="last_deactivation_run",
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



from datetime import datetime, timedelta

import pytz
from django.core.exceptions import ValidationError
from django.test import TestCase

from django_ecommerce.models import Category, Product, Color, Gender, ProductVariant, Warehouse, Discount, \
    DiscountCategory, \
    DiscountProduct, Stock, StockImport, Order, OrderItem, generate_order_code, Setting
from django_ecommerce.utils import money


class ModelsTestCase(TestCase):
    def setUp(self):
        product = Product.objects.create(
            name='Product1',
            category=Category.objects.create(name='Test'),
            color=Color.objects.create(name='Blue'),
            gender=Gender.objects.create(name='Unisex'),
            price=money(10)
        )

        ProductVariant.objects.create(product=product, name='Small', sku='PRODUCT1-SMALL')
        ProductVariant.objects.create(product=product, name='Medium', sku='PRODUCT1-MEDIUM')

        Warehouse.objects.create(name='Warehouse1')

    def test_category_discount_percentage(self):
        category = Category.objects.first()

        # No discount
        self.assertEqual(category.discount_percentage(), 0)

        # Create discount
        discount = Discount.objects.create(
            name='Discount1',
            start_date=datetime.now(tz=pytz.UTC),
            end_date=datetime.now(tz=pytz.UTC) + timedelta(days=7)
        )

        DiscountCategory.objects.create(discount=discount, category=category, percentage=10)

        self.assertEqual(category.discount_percentage(), 0)

        discount.active = True
        discount.save()

        self.assertEqual(category.discount_percentage(), 10)

    def test_product_discount_percentage(self):
        product = Product.objects.first()

        # No discount
        self.assertEqual(product.discount_percentage(), 0)

        # Create category discount
        discount = Discount.objects.create(
            name='Discount2',
            start_date=datetime.now(tz=pytz.UTC),
            end_date=datetime.now(tz=pytz.UTC) + timedelta(days=7),
            active=True
        )
        DiscountCategory.objects.create(discount=discount, category=product.category, percentage=20)
        self.assertEqual(product.discount_percentage(), 20)

        # Create product discount
        DiscountProduct.objects.create(discount=discount, product=product, percentage=30)
        self.assertEqual(product.discount_percentage(), 30)

    def test_product_get_actual_price(self):
        product = Product.objects.first()

        self.assertEqual(product.get_actual_price(), product.price)

        discount = Discount.objects.create(
            name='Discount3',
            start_date=datetime.now(tz=pytz.UTC),
            end_date=datetime.now(tz=pytz.UTC) + timedelta(days=7),
            active=True
        )

        DiscountProduct.objects.create(discount=discount, product=product, percentage=10)

        self.assertEqual(Product.objects.first().get_actual_price(), product.price * (100 - 10) / 100)

    def product_get_is_on_stock(self):
        product = Product.objects.first()

        self.assertFalse(product.get_is_on_stock())

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=product.variants.first(), warehouse=Warehouse.objects.first(), quantity=5,
                             content_object=si)

        self.assertTrue(product.get_is_on_stock())

    def test_product_variant_manager(self):
        product = Product.objects.first()

        # No stock
        variant = product.variants.first()
        self.assertEqual(variant.stock_quantity, 0)
        self.assertEqual(variant.reserved_quantity, 0)
        self.assertEqual(variant.available_quantity, 0)

        # Add stock
        si = StockImport.objects.create()
        Stock.objects.create(product_variant=product.variants.first(), warehouse=Warehouse.objects.first(), quantity=5,
                             content_object=si)
        variant = product.variants.first()
        self.assertEqual(variant.stock_quantity, 5)
        self.assertEqual(variant.reserved_quantity, 0)
        self.assertEqual(variant.available_quantity, 5)

        # Create order and reserve items
        order = Order.objects.create()
        OrderItem.objects.create(order=order, product_variant=variant, quantity=3, price=product.price, reserved=True)

        variant = product.variants.first()
        self.assertEqual(variant.stock_quantity, 5)
        self.assertEqual(variant.reserved_quantity, 3)
        self.assertEqual(variant.available_quantity, 2)

    def test_product_variant_stock_data(self):
        product = Product.objects.first()
        warehouse = Warehouse.objects.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=product.variants.first(), warehouse=warehouse, quantity=5,
                             content_object=si)

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=product.variants.first(), warehouse=warehouse, quantity=10,
                             content_object=si)

        self.assertEqual(product.variants.first().stock_data(), [(warehouse.name, 15)])

    def test_product_variant_reservations_data(self):
        product = Product.objects.first()
        variant = product.variants.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=variant, warehouse=Warehouse.objects.first(), quantity=5,
                             content_object=si)

        order = Order.objects.create()
        OrderItem.objects.create(order=order, product_variant=variant, quantity=3, price=product.price, reserved=True)

        self.assertEqual(variant.reservations_data(), [(order.code, 3)])

    def test_generate_order_code(self):
        code = generate_order_code()
        self.assertTrue(len(code) == 6)

    def test_order_total_amount(self):
        product = Product.objects.first()
        variant = product.variants.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=variant, warehouse=Warehouse.objects.first(), quantity=3,
                             content_object=si)

        order = Order.objects.create()
        OrderItem.objects.create(order=order, product_variant=variant, quantity=3, price=product.price, reserved=True)

        self.assertEqual(order.total_amount(), product.price * 3)

        # Enable shipping
        ecommerce_settings = Setting.get_solo()
        ecommerce_settings.shipping_enabled = True
        ecommerce_settings.shipping_amount = money(3)
        ecommerce_settings.save()

        order = Order.objects.create(shipping_amount=ecommerce_settings.shipping_amount)
        OrderItem.objects.create(order=order, product_variant=variant, quantity=3, price=product.price, reserved=True)

        self.assertEqual(order.total_amount(), product.price * 3 + ecommerce_settings.shipping_amount)

    def test_order_total_quantity(self):
        product = Product.objects.first()
        variant = product.variants.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=variant, warehouse=Warehouse.objects.first(), quantity=2,
                             content_object=si)

        order = Order.objects.create()
        OrderItem.objects.create(order=order, product_variant=variant, quantity=1, price=product.price, reserved=True)
        OrderItem.objects.create(order=order, product_variant=variant, quantity=1, price=product.price, reserved=True)

        self.assertEqual(order.total_quantity(), 2)

    def test_order_require_reserved_items(self):
        for status in Order.REQUIRE_RESERVED_ITEMS_STATUSES:
            order = Order.objects.create(status=status)
            self.assertTrue(order.require_reserved_items())

        order = Order.objects.create(status=Order.RETURN_TO_STOCK_STATUS)
        self.assertFalse(order.require_reserved_items())

    def test_order_can_modify_order_items(self):
        order = Order()
        self.assertTrue(order.can_modify_order_items())

        for status in Order.CAN_MODIFY_ORDER_ITEMS_STATUSES:
            order = Order.objects.create(status=status)
            self.assertTrue(order.can_modify_order_items())

        order = Order.objects.create(status=Order.RETURN_TO_STOCK_STATUS)
        self.assertFalse(order.can_modify_order_items())

    def test_order_status_changed(self):
        order = Order()
        order.status = Order.RETURN_TO_STOCK_STATUS
        self.assertTrue(order.status_changed())

    def test_order_check_status(self):
        for status in Order.ORDER_STATUSES:
            order = Order(status=status[0])
            allowed_statuses = Order.ALLOWED_STATUS_CHANGES[status[0]]

            for allowed in allowed_statuses:
                order.status = allowed
                try:
                    order.check_status()
                except ValidationError:
                    self.fail('ValidationError should not be raised')

    def test_order_remove_reservations(self):
        product = Product.objects.first()
        variant = product.variants.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=variant, warehouse=Warehouse.objects.first(), content_object=si,
                             quantity=5)

        order = Order.objects.create()
        OrderItem.objects.create(order=order, product_variant=variant, quantity=1, price=product.price, reserved=True)
        OrderItem.objects.create(order=order, product_variant=variant, quantity=1, price=product.price, reserved=True)

        order.remove_reservations()

        for item in order.items.all():
            self.assertFalse(item.reserved)

    def test_order_return_to_stock(self):
        product = Product.objects.first()
        variant = product.variants.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=variant, warehouse=Warehouse.objects.first(), content_object=si,
                             quantity=1)

        order = Order.objects.create()
        OrderItem.objects.create(
            order=order, product_variant=variant, quantity=1, price=product.price, reserved=True,
            warehouse=Warehouse.objects.first()
        )

        order.remove_from_stock()
        self.assertEqual(product.variants.first().stock_quantity, 0)

        order.return_to_stock()
        self.assertEqual(product.variants.first().stock_quantity, 1)

    def test_order_remove_from_stock(self):
        product = Product.objects.first()
        variant = product.variants.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=variant, warehouse=Warehouse.objects.first(), content_object=si,
                             quantity=2)

        order = Order.objects.create()
        OrderItem.objects.create(
            order=order, product_variant=variant, quantity=2, price=product.price, reserved=True,
            warehouse=Warehouse.objects.first()
        )

        order.remove_from_stock()
        self.assertEqual(product.variants.first().stock_quantity, 0)

    def test_order_should_remove_from_stock(self):
        order = Order(status=Order.REMOVE_FROM_STOCK_STATUS)
        self.assertTrue(order.should_remove_from_stock())

    def test_order_should_return_to_stock(self):
        order = Order(status=Order.RETURN_TO_STOCK_STATUS)
        self.assertTrue(order.should_return_to_stock())

    def test_order_should_remove_reservations(self):
        for status in Order.REMOVE_RESERVATIONS_STATUSES:
            order = Order(status=status)
            self.assertTrue(order.should_remove_reservations())

    def test_order_save(self):
        product = Product.objects.first()
        variant = product.variants.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=variant, warehouse=Warehouse.objects.first(), content_object=si,
                             quantity=3)

        order = Order.objects.create()
        OrderItem.objects.create(
            order=order, product_variant=variant, quantity=3, price=product.price, reserved=True,
            warehouse=Warehouse.objects.first()
        )

        order.status = Order.REMOVE_FROM_STOCK_STATUS
        order.save()

        self.assertEqual(product.variants.first().stock_quantity, 0)

        order.status = Order.RETURN_TO_STOCK_STATUS
        order.save()

        self.assertEqual(product.variants.first().stock_quantity, 3)

    def test_order_item_total(self):
        product = Product.objects.first()
        variant = product.variants.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=variant, warehouse=Warehouse.objects.first(), quantity=10,
                             content_object=si)

        order = Order.objects.create()
        item = OrderItem.objects.create(order=order, product_variant=variant, quantity=5, price=product.price,
                                        reserved=True)

        self.assertEqual(item.total(), 5 * product.price)

    def test_order_item_check_available_quantity(self):
        product = Product.objects.first()
        variant = product.variants.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=variant, warehouse=Warehouse.objects.first(), quantity=1,
                             content_object=si)

        order = Order.objects.create()

        with self.assertRaises(ValidationError):
            item = OrderItem(order=order, product_variant=variant, quantity=5, price=product.price,
                             reserved=True, warehouse=Warehouse.objects.first())
            item.check_available_quantity()

    def test_order_item_check_reserved_items(self):
        product = Product.objects.first()
        variant = product.variants.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=variant, quantity=3, warehouse=Warehouse.objects.first(),
                             content_object=si)

        order = Order.objects.create(status=Order.REQUIRE_RESERVED_ITEMS_STATUSES[0])

        with self.assertRaises(ValidationError):
            item = OrderItem(order=order, product_variant=variant, quantity=3, price=product.price,
                             reserved=False, warehouse=Warehouse.objects.first())
            item.check_reserved_items()

    def test_order_item_get_actual_price(self):
        product = Product.objects.first()
        variant = product.variants.first()

        si = StockImport.objects.create()
        Stock.objects.create(product_variant=variant, warehouse=Warehouse.objects.first(), content_object=si,
                             quantity=4)

        order = Order.objects.create()
        item = OrderItem(order=order, product_variant=variant, quantity=3, price=product.price, discount_percentage=10,
                         reserved=False, warehouse=Warehouse.objects.first())

        self.assertEqual(item.get_actual_price(), item.price * (100 - 10) / 100)



from django.core.cache import cache
from django.core.management.base import BaseCommand
from django.utils import timezone
from django.db.models import F
from django_ecommerce.models import Discount


class Command(BaseCommand):
    def handle(self, *args, **kwargs):
        now = timezone.now()

        # ACTIVATIONS
        to_activate = Discount.objects.filter(
            start_date__lte=now,
            end_date__gt=now,
        ).exclude(last_activation_run__gte=F("start_date"))

        for d in to_activate:
            self.invalidate_for_discount(d)
            d.last_activation_run = now
            d.save(update_fields=["last_activation_run"])

        # DEACTIVATIONS
        to_deactivate = Discount.objects.filter(end_date__lte=now).exclude(
            last_deactivation_run__gte=F("end_date")
        )

        for d in to_deactivate:
            self.invalidate_for_discount(d)
            d.last_deactivation_run = now
            d.save(update_fields=["last_deactivation_run"])

    def invalidate_for_discount(self, discount):
        # Todo: invalidate cache for products and categories affected by the discount
        cache.clear()



from datetime import datetime

import pytz
from autoslug import AutoSlugField
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Sum, F, Value, Subquery, OuterRef, IntegerField
from django.db.models.functions import Coalesce
from django.urls import reverse
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.safestring import mark_safe
from django_cloud_storage.storage import CloudStorage
from django_cloud_thumbnails import ImageRatioField, ImageCropField
from djmoney.models.fields import MoneyField
from solo.models import SingletonModel


class Category(models.Model):
    name = models.CharField(max_length=255)
    slug = AutoSlugField(populate_from="name", editable=True, blank=True, unique=True)
    description = models.TextField(blank=True, null=True)
    seo_title = models.CharField(max_length=60, blank=True, null=True)
    seo_description = models.TextField(max_length=160, blank=True, null=True)
    seo_keywords = models.TextField(max_length=255, blank=True, null=True)
    seo_image = ImageCropField(blank=True, null=True, storage=CloudStorage)
    seo_image_cropping = ImageRatioField("seo_image", "1200x630")
    seo_image_cropped = models.ImageField(
        storage=CloudStorage, blank=True, null=True, editable=False
    )

    def discount_percentage(self):
        if not self.pk:
            return 0

        now = datetime.now(pytz.timezone(settings.TIME_ZONE))

        discount = self.discount_categories.filter(
            discount__active=True,
            discount__start_date__lte=now,
            discount__end_date__gte=now,
        ).first()

        if discount:
            return discount.percentage

        return 0

    def __str__(self):
        return self.name

    class Meta:
        verbose_name_plural = "categories"


class Color(models.Model):
    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


class City(models.Model):
    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name_plural = "cities"


class Country(models.Model):
    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name

    class Meta:
        verbose_name_plural = "countries"


class Gender(models.Model):
    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


class Product(models.Model):
    name = models.CharField(max_length=255)
    slug = AutoSlugField(populate_from="name", editable=True, blank=True, unique=True)
    description = models.TextField(null=True, blank=True)
    category = models.ForeignKey(
        Category, related_name="products", null=True, on_delete=models.SET_NULL
    )
    color = models.ForeignKey(Color, null=True, on_delete=models.SET_NULL)
    gender = models.ForeignKey(Gender, null=True, on_delete=models.SET_NULL)
    price = MoneyField(
        max_digits=14,
        decimal_places=2,
        default_currency=settings.ECOMMERCE_DEFAULT_CURRENCY,
        currency_choices=settings.ECOMMERCE_AVAILABLE_CURRENCIES,
    )
    seo_title = models.CharField(max_length=60, blank=True, null=True)
    seo_description = models.TextField(max_length=160, blank=True, null=True)
    seo_keywords = models.TextField(max_length=255, blank=True, null=True)
    seo_image = ImageCropField(blank=True, null=True, storage=CloudStorage)
    seo_image_cropping = ImageRatioField("seo_image", "1200x630")
    seo_image_cropped = models.ImageField(
        storage=CloudStorage, blank=True, null=True, editable=False
    )

    def __str__(self):
        return self.name

    def __init__(self, *args, **kwargs):
        super(Product, self).__init__(*args, **kwargs)

        self.discount = self.discount_percentage()
        self.has_discount = self.discount > 0
        self.actual_price = self.get_actual_price()
        self.is_on_stock = self.get_is_on_stock()

    def discount_percentage(self):
        now = datetime.now(pytz.timezone(settings.TIME_ZONE))

        if self.pk:
            discount = self.discount_products.filter(
                discount__active=True,
                discount__start_date__lte=now,
                discount__end_date__gte=now,
            ).first()

            if discount:
                return discount.percentage

        if self.category:
            return self.category.discount_percentage()

        return 0

    def get_actual_price(self):
        if not self.price:
            return 0

        return self.price * (100 - self.discount) / 100

    def get_is_on_stock(self):
        if self.pk:
            for variant in self.variants.all():
                if variant.available_quantity > 0:
                    return True

        return False


class ProductVariantManager(models.Manager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                stock_quantity=Coalesce(
                    Subquery(
                        Stock.objects.filter(product_variant=OuterRef("pk"))
                        .values("product_variant")
                        .annotate(total_stock=Sum("quantity"))
                        .values("total_stock")[:1]
                    ),
                    Value(0),
                    output_field=IntegerField(),
                ),
                reserved_quantity=Coalesce(
                    Subquery(
                        OrderItem.objects.filter(
                            product_variant=OuterRef("pk"), reserved=True
                        )
                        .values("product_variant")
                        .annotate(total_reserved=Sum("quantity"))
                        .values("total_reserved")[:1]
                    ),
                    Value(0),
                    output_field=IntegerField(),
                ),
            )
            .annotate(available_quantity=F("stock_quantity") - F("reserved_quantity"))
        )


class ProductVariant(models.Model):
    product = models.ForeignKey(
        Product, related_name="variants", on_delete=models.CASCADE
    )
    sku = models.CharField(max_length=255, unique=True)
    name = models.CharField(max_length=255)
    position = models.PositiveIntegerField(default=0)
    active = models.BooleanField(default=True)

    objects = ProductVariantManager()

    reserved_quantity = 0
    stock_quantity = 0
    available_quantity = 0

    def save(self, *args, **kwargs):
        self.sku = self.sku.upper()

        super().save(*args, **kwargs)

    def stock_data(self) -> list[tuple[str, int]]:
        result = []

        rows = self.stock.values("warehouse_id").annotate(total=Sum("quantity"))

        for row in rows:
            warehouse = Warehouse.objects.get(pk=row["warehouse_id"])
            result.append((warehouse.name, row["total"]))

        return result

    def reservations_data(self) -> list[tuple[str, int]]:
        result = []

        for item in self.order_items.filter(reserved=True).all():
            result.append((item.order.code, item.quantity))

        return result

    def __str__(self):
        return self.sku


class ProductImage(models.Model):
    product = models.ForeignKey(
        Product, related_name="images", on_delete=models.CASCADE
    )
    image = ImageCropField(blank=True, null=True, storage=CloudStorage)
    cropping = ImageRatioField("image", "1000x1000")
    cropped = models.ImageField(
        storage=CloudStorage, blank=True, null=True, editable=False
    )
    position = models.PositiveIntegerField(default=0, blank=False, null=False)

    class Meta:
        ordering = ["position"]

    def __str__(self):
        return f"{self.product} ({self.position})"


class Warehouse(models.Model):
    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


class Stock(models.Model):
    product_variant = models.ForeignKey(
        ProductVariant, related_name="stock", on_delete=models.CASCADE
    )
    warehouse = models.ForeignKey(
        Warehouse, related_name="stock", on_delete=models.CASCADE
    )
    quantity = models.IntegerField()
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, null=True, blank=True
    )
    object_id = models.PositiveIntegerField(null=True, blank=True)
    content_object = GenericForeignKey("content_type", "object_id")
    created_at = models.DateTimeField(auto_now_add=True)

    def content_object_class(self):
        if self.content_type:
            return self.content_type.model_class()._meta.verbose_name.capitalize()

    def content_object_link(self):
        if self.content_type:
            url = reverse(
                f"admin:{self.content_type.app_label}_{self.content_type.model}_change",
                args=[self.object_id],
            )

            return mark_safe(f'<a href="{url}">{self.content_object_class()}</a>')

    content_object_link.short_description = "Source"

    class Meta:
        verbose_name_plural = "stock"

        indexes = [
            models.Index(fields=["content_type", "object_id"]),
        ]

    def __str__(self):
        return f"{self.product_variant.product.name} ({self.warehouse.name}) ({self.quantity})"


class StockImport(models.Model):
    comment = models.TextField(blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True)
    items = GenericRelation(Stock)

    def __str__(self):
        tz = timezone.get_current_timezone()

        return self.created_at.astimezone(tz).strftime("%d/%m/%Y %H:%M")


class Setting(SingletonModel):
    shipping_enabled = models.BooleanField(default=False)
    shipping_amount = MoneyField(
        max_digits=14,
        decimal_places=2,
        default=0,
        null=True,
        blank=True,
        default_currency=settings.ECOMMERCE_DEFAULT_CURRENCY,
        currency_choices=settings.ECOMMERCE_AVAILABLE_CURRENCIES,
    )
    free_shipping_enabled = models.BooleanField(default=False)
    free_shipping_min_amount = MoneyField(
        max_digits=14,
        decimal_places=2,
        default=0,
        blank=True,
        null=True,
        default_currency=settings.ECOMMERCE_DEFAULT_CURRENCY,
        currency_choices=settings.ECOMMERCE_AVAILABLE_CURRENCIES,
    )

    def __str__(self):
        return "Settings"

    class Meta:
        verbose_name = "Settings"


def generate_order_code():
    code = get_random_string(6).upper()

    try:
        Order.objects.get(code=code)
        return generate_order_code()
    except Order.DoesNotExist:
        return code


class Order(models.Model):
    ORDER_STATUSES = (
        ("pending", "Pending"),
        ("confirmed", "Confirmed"),
        ("canceled", "Canceled"),
        ("sent", "Sent"),
        ("returned", "Returned"),
        ("delivered", "Delivered"),
    )

    ALLOWED_STATUS_CHANGES = {
        "pending": ("confirmed", "canceled", "sent", "delivered"),
        "confirmed": ("canceled", "sent", "delivered"),
        "canceled": (),
        "sent": ("returned",),
        "returned": (),
        "delivered": ("returned",),
    }

    REMOVE_FROM_STOCK_STATUS = "delivered"

    RETURN_TO_STOCK_STATUS = "returned"

    REQUIRE_RESERVED_ITEMS_STATUSES = ("confirmed", "sent")

    REMOVE_RESERVATIONS_STATUSES = ("canceled", "returned", "delivered")

    CAN_MODIFY_ORDER_ITEMS_STATUSES = ("pending", "confirmed", "sent")

    code = models.CharField(
        max_length=16, blank=True, default=generate_order_code, db_index=True
    )
    status = models.CharField(
        max_length=16, choices=ORDER_STATUSES, default=ORDER_STATUSES[0][0]
    )
    email = models.EmailField(max_length=255, null=True, blank=True)
    full_name = models.CharField(max_length=255, null=True, blank=True)
    address = models.CharField(max_length=255, null=True, blank=True)
    phone = models.CharField(max_length=255, null=True, blank=True)
    city = models.CharField(max_length=255, null=True, blank=True)
    country = models.CharField(max_length=255, null=True, blank=True)
    zip_code = models.CharField(max_length=255, null=True, blank=True)
    remark = models.TextField(blank=True, null=True)
    shipping_amount = MoneyField(
        max_digits=14,
        decimal_places=2,
        blank=True,
        null=True,
        default=0,
        default_currency=settings.ECOMMERCE_DEFAULT_CURRENCY,
        currency_choices=settings.ECOMMERCE_AVAILABLE_CURRENCIES,
    )
    created_at = models.DateTimeField(auto_now_add=True)

    __original_status = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__original_status = self.status

    def total_amount(self):
        total = self.shipping_amount

        for item in self.items.all():
            total += item.actual_price * item.quantity

        return total

    def total_quantity(self):
        quantity = 0

        for item in self.items.all():
            quantity += item.quantity

        return quantity

    def require_reserved_items(self):
        return self.status in self.REQUIRE_RESERVED_ITEMS_STATUSES

    def can_modify_order_items(self):
        if not self.pk:
            return True

        return self.status in self.CAN_MODIFY_ORDER_ITEMS_STATUSES

    def status_changed(self):
        return self.__original_status != self.status

    def check_status(self):
        if self.status_changed() and self.__original_status:
            allowed_statuses = self.ALLOWED_STATUS_CHANGES[self.__original_status]

            if not allowed_statuses:
                raise ValidationError(
                    f"Status of order with  status {self.__original_status} cannot be changed."
                )

            if self.status not in allowed_statuses:
                statuses = ", ".join(allowed_statuses)
                raise ValidationError(
                    f"Order with status {self.__original_status} can only be changed to {statuses}"
                )

    def remove_reservations(self):
        for item in self.items.all():
            item.reserved = False
            item.save()

    def return_to_stock(self):
        for item in self.items.all():
            stock = Stock(
                product_variant=item.product_variant,
                warehouse=item.warehouse,
                quantity=item.quantity,
                content_object=self,
            )
            stock.save()

    def remove_from_stock(self):
        for item in self.items.all():
            stock = Stock(
                product_variant=item.product_variant,
                warehouse=item.warehouse,
                quantity=-item.quantity,
                content_object=self,
            )
            stock.save()

    def clean(self):
        self.check_status()

    def should_remove_from_stock(self):
        return self.status == self.REMOVE_FROM_STOCK_STATUS

    def should_return_to_stock(self):
        return self.status == self.RETURN_TO_STOCK_STATUS

    def should_remove_reservations(self):
        return self.status in self.REMOVE_RESERVATIONS_STATUSES

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

        if self.status_changed():
            if self.should_return_to_stock():
                self.return_to_stock()
            elif self.should_remove_from_stock():
                self.remove_from_stock()

            if self.should_remove_reservations():
                self.remove_reservations()

    def __str__(self):
        return self.code


class OrderItem(models.Model):
    order = models.ForeignKey(Order, related_name="items", on_delete=models.CASCADE)
    product_variant = models.ForeignKey(
        ProductVariant, related_name="order_items", on_delete=models.CASCADE
    )
    price = MoneyField(
        max_digits=14,
        decimal_places=2,
        default_currency=settings.ECOMMERCE_DEFAULT_CURRENCY,
        currency_choices=settings.ECOMMERCE_AVAILABLE_CURRENCIES,
    )
    discount_percentage = models.PositiveIntegerField(default=0, blank=True, null=True)
    quantity = models.PositiveIntegerField()
    warehouse = models.ForeignKey(
        Warehouse, related_name="order_items", on_delete=models.SET_NULL, null=True
    )
    reserved = models.BooleanField(default=False)

    def total(self):
        return self.actual_price * self.quantity

    def check_available_quantity(self):
        if self.warehouse:

            try:
                total_quantity = self.quantity

                items = OrderItem.objects.filter(
                    warehouse=self.warehouse,
                    product_variant=self.product_variant,
                    reserved=True,
                ).all()

                for item in items:
                    if item.pk == self.pk:
                        continue

                    total_quantity += item.quantity

                self.product_variant.stock.filter(
                    quantity__gte=total_quantity, warehouse=self.warehouse
                ).get()
            except Stock.DoesNotExist:
                raise ValidationError(
                    "Selected warehouse does not have enough quantity of this item on stock."
                )

    def __init__(self, *args, **kwargs):
        super(OrderItem, self).__init__(*args, **kwargs)

        self.has_discount = self.discount_percentage > 0
        self.actual_price = self.get_actual_price()

    def check_reserved_items(self):
        if self.order.require_reserved_items() and not self.reserved:
            raise ValidationError(
                f"Order items must be reserved to change the status to {self.order.status}"
            )

    def clean(self):
        self.check_reserved_items()
        self.check_available_quantity()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

    def get_actual_price(self):
        if not self.price:
            return 0

        return self.price * (100 - int(self.discount_percentage)) / 100

    def __str__(self):
        return "Order item"


class Discount(models.Model):
    active = models.BooleanField(default=False)
    name = models.CharField(max_length=255)
    start_date = models.DateTimeField()
    end_date = models.DateTimeField()
    last_activation_run = models.DateTimeField(null=True, blank=True)
    last_deactivation_run = models.DateTimeField(null=True, blank=True)

    def __str__(self):
        return self.name


class DiscountCategory(models.Model):
    discount = models.ForeignKey(
        Discount, related_name="categories", on_delete=models.CASCADE
    )
    category = models.ForeignKey(
        Category, related_name="discount_categories", on_delete=models.CASCADE
    )
    percentage = models.PositiveIntegerField(default=0)

    class Meta:
        verbose_name_plural = "discount categories"

    def __str__(self):
        return self.category.name


class DiscountProduct(models.Model):
    discount = models.ForeignKey(
        Discount, related_name="products", on_delete=models.CASCADE
    )
    product = models.ForeignKey(
        Product, related_name="discount_products", on_delete=models.CASCADE
    )
    percentage = models.PositiveIntegerField(default=0)

    def __str__(self):
        return self.product.name
